
#include <math.h>

#include <gmm/gmm.h>

#include <Alignment.h>

namespace FriGeo
{

void
Alignment::ComputeT (const std::vector<double>& p, gmm::dense_matrix<double>& Tr) const
{
	double t_x = p[0];
	double t_y = p[1];
	double t_z = p[2];
	double alpha = p[3];
	double beta = p[4];
	double gamma = p[5];

	/* from http://planning.cs.uiuc.edu/node104.html
	 */
	gmm::resize (Tr, 4, 4);

	Tr(0,0) = cos(alpha)*cos(beta);
	Tr(0,1) = cos(alpha)*sin(beta)*sin(gamma)-sin(alpha)*cos(gamma);
	Tr(0,2) = cos(alpha)*sin(beta)*cos(gamma)+sin(alpha)*sin(gamma);
	Tr(0,3) = t_x;

	Tr(1,0) = sin(alpha)*cos(beta);
	Tr(1,1) = sin(alpha)*sin(beta)*sin(gamma)+cos(alpha)*cos(gamma);
	Tr(1,2) = sin(alpha)*sin(beta)*cos(gamma)-cos(alpha)*sin(gamma);
	Tr(1,3) = t_y;

	Tr(2,0) = -sin(beta);
	Tr(2,1) = cos(beta)*sin(gamma);
	Tr(2,2) = cos(beta)*cos(gamma);
	Tr(2,3) = t_z;

	Tr(3,0) = Tr(3,1) = Tr(3,2) = 0;
	Tr(3,3) = 1;
}

double
Alignment::FunctionValue (const std::vector<double>& p) const
{
	return (FunctionValue (p, NULL));
}

double
Alignment::FunctionValue (const std::vector<double>& p,
	std::vector<double>* individualLoss) const
{
	gmm::dense_matrix<double> Tr;
	ComputeT (p, Tr);

	/* Perform transformation, then squared-deviation objective.
	 */
	gmm::dense_matrix<double> TX (gmm::mat_nrows (X), gmm::mat_ncols (X));
	gmm::mult (X, gmm::transposed (Tr), TX);
	gmm::add (gmm::scaled (U, -1.0), TX, TX);

	if (individualLoss != NULL) {
		individualLoss->resize (gmm::mat_nrows (X));
		gmm::clear (*individualLoss);
	}

	double fval = 0.0;
	for (unsigned int n = 0 ; n < gmm::mat_nrows (TX) ; ++n) {
		for (unsigned int c = 0 ; c < gmm::mat_ncols (TX) ; ++c) {
			double loss = TX(n,c)*TX(n,c);

			fval += loss;
			if (individualLoss != NULL)
				(*individualLoss)[n] += loss;
		}
	}

	return (fval);
}

void
Alignment::Gradient (const std::vector<double>& p, std::vector<double>& grad) const
{
	double alpha = p[3];
	double beta = p[4];
	double gamma = p[5];

	gmm::dense_matrix<double> Malpha (4, 4);
	Malpha(0,0) = -sin(alpha)*cos(beta);
	Malpha(0,1) = -sin(alpha)*sin(beta)*sin(gamma)-cos(alpha)*cos(gamma);
	Malpha(0,2) = -sin(alpha)*sin(beta)*cos(gamma)+cos(alpha)*sin(gamma);
	Malpha(0,3) = 0;

	Malpha(1,0) = cos(alpha)*cos(beta);
	Malpha(1,1) = cos(alpha)*sin(beta)*sin(gamma)-sin(alpha)*cos(gamma);
	Malpha(1,2) = cos(alpha)*sin(beta)*cos(gamma)+sin(alpha)*sin(gamma);
	Malpha(1,3) = 0;

	Malpha(2,0) = Malpha(2,1) = Malpha(2,2) = Malpha(2,3) = 0;
	Malpha(3,0) = Malpha(3,1) = Malpha(3,2) = Malpha(3,3) = 0;

	gmm::dense_matrix<double> Mbeta (4, 4);
	Mbeta(0,0) = -cos(alpha)*sin(beta);
	Mbeta(0,1) = cos(alpha)*cos(beta)*sin(gamma);
	Mbeta(0,2) = cos(alpha)*cos(beta)*cos(gamma);
	Mbeta(0,3) = 0;

	Mbeta(1,0) = -sin(alpha)*sin(beta);
	Mbeta(1,1) = sin(alpha)*cos(beta)*sin(gamma);
	Mbeta(1,2) = sin(alpha)*cos(beta)*cos(gamma);
	Mbeta(1,3) = 0;

	Mbeta(2,0) = -cos(beta);
	Mbeta(2,1) = -sin(beta)*sin(gamma);
	Mbeta(2,2) = -sin(beta)*cos(gamma);
	Mbeta(2,3) = 0;

	Mbeta(3,0) = Mbeta(3,1) = Mbeta(3,2) = Mbeta(3,3) = 0;

	gmm::dense_matrix<double> Mgamma (4, 4);
	Mgamma(0,0) = 0;
	Mgamma(0,1) = cos(alpha)*sin(beta)*cos(gamma)+sin(alpha)*sin(gamma);
	Mgamma(0,2) = -cos(alpha)*sin(beta)*sin(gamma)+sin(alpha)*cos(gamma);
	Mgamma(0,3) = 0;

	Mgamma(1,0) = 0;
	Mgamma(1,1) = sin(alpha)*sin(beta)*cos(gamma)-cos(alpha)*sin(gamma);
	Mgamma(1,2) = -sin(alpha)*sin(beta)*sin(gamma)-cos(alpha)*cos(gamma);
	Mgamma(1,3) = 0;

	Mgamma(2,0) = 0;
	Mgamma(2,1) = cos(beta)*cos(gamma);
	Mgamma(2,2) = -cos(beta)*sin(gamma);
	Mgamma(2,3) = 0;

	Mgamma(3,0) = Mgamma(3,1) = Mgamma(3,2) = Mgamma(3,3) = 0;

	gmm::dense_matrix<double> Tr;
	ComputeT (p, Tr);

	/* Perform transformation, then squared-deviation objective.
	 */
	gmm::dense_matrix<double> TX (gmm::mat_nrows (X), gmm::mat_ncols (X));
	gmm::mult (X, gmm::transposed (Tr), TX);
	gmm::add (gmm::scaled (U, -1.0), TX, TX);

	/* Gradients with respect to tx, ty, tz.
	 */
	gmm::resize (grad, 6);
	gmm::clear (grad);

	for (unsigned int n = 0 ; n < gmm::mat_nrows (TX) ; ++n)
		for (unsigned int c = 0 ; c < (gmm::mat_ncols (TX) - 1) ; ++c)
			grad[c] += 2.0*TX(n,c);

	/* Gradients with respect to alpha, beta, gamma
	 */
	gmm::dense_matrix<double> MaX (gmm::mat_nrows (X), gmm::mat_ncols (X));
	gmm::dense_matrix<double> MbX (gmm::mat_nrows (X), gmm::mat_ncols (X));
	gmm::dense_matrix<double> MgX (gmm::mat_nrows (X), gmm::mat_ncols (X));

	gmm::mult (X, gmm::transposed (Malpha), MaX);
	gmm::mult (X, gmm::transposed (Mbeta), MbX);
	gmm::mult (X, gmm::transposed (Mgamma), MgX);

	for (unsigned int n = 0 ; n < gmm::mat_nrows (TX) ; ++n) {
		for (unsigned int c = 0 ; c < gmm::mat_ncols (TX) ; ++c) {
			grad[3] += 2.0*TX(n,c)*MaX(n,c);
			grad[4] += 2.0*TX(n,c)*MbX(n,c);
			grad[5] += 2.0*TX(n,c)*MgX(n,c);
		}
	}
}


double
Alignment::FindLeastSquaresAlignment (std::vector<double>* individualLoss)
{
	std::vector<double> p(6);
	p[0] = p[1] = p[2] = p[3] = p[4] = p[5] = 1;

	gmm::iteration iter(1e-8);
	gmm::bfgs (CallerFunctionValue (this), CallerGradient (this), p, 100, iter);

	ComputeT (p, T);

	double fval = FunctionValue (p, individualLoss);

	return (fval);
}

}

