
#ifndef	ALIGNMENT_H
#define	ALIGNMENT_H

/* Find the rigid 3D homogeneous transformation matrix (translation +
 * rotation) that aligns two set of points such that the minimum squared
 * distance between each point pair is minimized.
 *
 * Author: Sebastian Nowozin <sebastian.nowozin@tuebingen.mpg.de>
 * Date: 13th August 2007
 */

#include <vector>
#include <assert.h>

#include <gmm/gmm.h>

namespace FriGeo
{

class
Alignment
{
private:
	/* (N,4) homogeneous 3D input (X) and output (U) coordinates
	 */
	const gmm::dense_matrix<double>& X;
	const gmm::dense_matrix<double>& U;

	void ComputeT (const std::vector<double>& p, gmm::dense_matrix<double>& T) const;

	double FunctionValue (const std::vector<double>& p) const;
	void Gradient (const std::vector<double>& p, std::vector<double>& grad) const;

	double FunctionValue (const std::vector<double>& p,
		std::vector<double>* individualLoss) const;

	struct CallerFunctionValue
	{
	private:
		Alignment* baseAlign;

	public:
		double operator()(const std::vector<double>& p) const
		{
			return (baseAlign->FunctionValue (p));
		}

		CallerFunctionValue (Alignment* baseAlign)
			: baseAlign (baseAlign)
		{
		}
	};

	struct CallerGradient
	{
	private:
		Alignment* baseAlign;

	public:
		void operator()(const std::vector<double>& p, std::vector<double>& grad) const
		{
			baseAlign->Gradient (p, grad);
		}

		CallerGradient (Alignment* baseAlign)
			: baseAlign (baseAlign)
		{
		}
	};


public:
	/* After calling FindLeastSquaresAlignment, the minimizing transformation
	 * matrix is available through T.  The transformation can be carried out
	 * using: U = X*T'.
	 */
	gmm::dense_matrix<double> T;

	/* X: (N,4) matrix of input space points.
	 * U: (N,4) matrix of output space points.
	 *
	 * The coordinates are homogeneous, such that the fourth column is set to
	 * one for each point.
	 */
	Alignment (const gmm::dense_matrix<double>& X, const gmm::dense_matrix<double>& U)
		: X (X), U (U)
	{
		assert (gmm::mat_nrows (X) == gmm::mat_nrows (U));
		assert (gmm::mat_ncols (X) == gmm::mat_ncols (U));
	}

	/* Find the minimum squared distance alignment between X and U.
	 *
	 * individualLoss: N-vector of individual squared distances.
	 *
	 * Return the sum of squared distances.
	 */
	double FindLeastSquaresAlignment (std::vector<double>* individualLoss);
};

};

#endif


