/* Frequent Geometric Subgraph Retrieval (FreqGeo).
 *
 * Copyright (C) 2008 -- Sebastian Nowozin <sebastian.nowozin@tuebingen.mpg.de>
 * This file is part of FreqGeo.
 *
 * FreqGeo is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * FreqGeo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with FreqGeo.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef	FREQGEO_H
#define	FREQGEO_H

#include <iostream>
#include <fstream>
#include <vector>

#include <Graph.h>

namespace FriGeo
{

class FreqGeo
{
private:
	mutable unsigned int graphCount;

	const std::vector<Graph>& G;
	const Graph& refg;
	unsigned int minsup;
	double vertexTolerance;

	/* Check conditions for subg being visited the first time:
	 *    i) it appears only once in L[refg], or
	 *   ii) it appears more than once but is the first occurence.
	 * Return true if it is the first time visit of this subgraph, false
	 * otherwise.  For this to work, refg_index must be >= 0.  If refg_index
	 * == -1, we always return true and uniqueness of the reported subgraphs
	 * cannot be guaranteed.
	 */
	bool IsFirstTimeVisitSubgraph (const Graph& subg,
		const std::vector<unsigned int>& PI,
		const std::map<unsigned int, std::vector<std::vector<unsigned int> > >& L) const;

	/* Recursive mining method.
	 *
	 * subg: The current subgraph being grown in refg.
	 * PI: The index mapping (PI.size() == gmm::mat_nrows(subg.VP)), such that
	 *     the n'th node in subg is the PI[n]'th node in refg.
	 * L: The occurence list, where .first is the index into 'G', and .second
	 *    is a vector of index mappings of nodes from subg into G[.first].
	 */
	void MineR (const Graph& subg, const std::vector<unsigned int>& PI,
		const std::map<unsigned int,
			std::vector<std::vector<unsigned int> > >& L);

protected:
	int refg_index;	// If -1 unknown, otherwise: refg = G[refg_index]

	/* Report a frequent connected geometric subgraph.
	 *
	 * subg: The graph which is frequent.  The support is given by L.size().
	 * L: The location list, reporting the complete set of mappings of subg
	 *    into the graph database.  L[gidx] is the list of index mappings for
	 *    the gidx'th subgraph in the graph database.
	 */
	virtual void Report (const Graph& subg,
		const std::map<unsigned int,
			std::vector<std::vector<unsigned int> > >& L);

public:
	/* Frequent geometric subgraph mining.
	 *
	 * G: The graph database.
	 * refg: The reference graph.
	 * minsup: Minimum support threshold in the graph database.
	 * vertexTolerance: max-tolerance on the squared vertex distances.
	 */
	FreqGeo (const std::vector<Graph>& G, const Graph& refg,
		unsigned int minsup, double vertexTolerance = 1e-2)
		: graphCount (0), G (G), refg (refg), refg_index(-1),
			minsup (minsup),
			vertexTolerance (vertexTolerance)
	{
	}

	// refg_index: Select a subgraph from G, 0 <= refg_index < G.size().
	FreqGeo (const std::vector<Graph>& G, unsigned int refg_index,
		unsigned int minsup, double vertexTolerance = 1e-2)
		: graphCount (0), G (G), refg (this->G[refg_index]),
			refg_index(refg_index), minsup (minsup),
			vertexTolerance (vertexTolerance)
	{
	}

	virtual ~FreqGeo (void)
	{
	}

	/* Enumerate all subgraphs of a reference graph 'refg'.  For each
	 * enumerated graph all possible mappings of refg into the graph database
	 * 'G' are considered and for sufficiently many mappings (> minsup) the
	 * subgraph is output.
	 */
	void Mine (void);
};

};

#endif


