function [subg, count, GY] = gspan (G, minsup, size_req, directed, refg_index);

% gSpan frequent graph substructure algorithm, retrieval version.
%
% (This is a Matlab wrapper around an extended version of the gspan
%  implementation by Taku Kudo.)
%
% Author: Sebastian Nowozin <sebastian.nowozin@tuebingen.mpg.de>
% Date: 7th December 2006, updated 16th April 2008
%
% Find all subgraphs in the set of graphs G that appear at least minsup times
% and are the subgraph of a given reference graph.
%
%%%
% Input
%   G: (1,n) cellarray of n graph structures with this layout
%     g.nodelabels: (n,1) discrete integer labels [L_1 ; L_2 ; ... ; L_n];
%     g.edges: (m,2) edges, [from to] at each line:
%       [e_1_{from} e_1_{to} edgelabel_1 ; ... ; e_m_{from} e_m_{to} edgelabel_m]
%       The node indices go from 1 to n.  (They will be converted to 0-(n-1)
%       internally and converted back.)
%
%   minsup: The minimum number of times a frequent pattern has to appear in G.
%
%   (optional) size_req: Either [] for no limit, or (1,2) real col-vector
%      [min_node_count max_node_count], which limits the returned graphs to
%      have a number of nodes p, such that
%         min_node_count <= p <= max_node_count.
%      If max_node_count < min_node_count, only the lower bound is used.
%      (default: [0 0])
%
%   (optional) directed: If 1, directed graphs are used, default
%      is 0: undirected graphs
%
%   (optional) refg_index: Between 1 and n the index of the reference graph.
%
%%%
% Output
%   subg: (1,p) cellarray of p graph structures that appear frequently in G.
%
%   (optional) count: (1,p) uint32 array giving at element i the number of times
%     graphs{i} appears in G (duplicates within one graph only counted once).
%
%   (optional) GY: (n,p) double array giving individual counts for each
%       subgraph/graph pair.  At (i,j) the number of counts of the j'th
%       subgraph found in the i'th graph is recorded.  This takes no
%       additional computation time to record.  For transductive samples, the
%       indicator function is returned.
%

if nargin < 2 || nargin > 5 || nargout > 3
	error(['Invalid number of input or output parameters.']);
end

d_size_req = [0 0];
if nargin >= 3
	if isempty(size_req)
		size_req = [0 0];
	end

	if size(size_req,1) ~= 1 || size(size_req,2) ~= 2
		error(['size_req parameter invalid.']);
	end
	d_size_req = size_req;
end

d_directed = 0;
if nargin >= 4
	d_directed = directed;
end

d_refg_index = -1;
if nargin >= 5
	d_refg_index = refg_index - 1;
end

for i=1:length(G)
	[v, r] = verifygraph (G{i});
	if v ~= 1
		error(['Graph ', num2str(i), ': ', r]);
	end
end
if size(G,2) == 1
	G=G';	% put it as row.
end

disp(['Starting normal gspan run...']);
if nargout == 3
	[subg, count, GY] = mexgspan (G, minsup, d_size_req, d_directed, d_refg_index);
elseif nargout == 2
	[subg, count] = mexgspan (G, minsup, d_size_req, d_directed, d_refg_index);
elseif nargout == 1
	[subg] = mexgspan (G, minsup, d_size_req, d_directed, d_refg_index);
end

% Filter out duplicates (this is due to undirected edges)
if d_directed == 0
	for i=1:length(subg)
		subg{i}.edges = unique (subg{i}.edges, 'rows');
	end
end


function [valid, reason] = verifygraph (G);

% Verify if the directed graph G is valid, that is:
%
% 1. Does not contain any self-referential or duplicate edges.
% 2. Uses proper indices.

nodes = size(G.nodelabels,1);

% Check for indices
if size(G.edges, 1) > 0
	I = find (G.edges(:,1) <= 0 | G.edges(:,1) > nodes | ...
		G.edges(:,2) <= 0 | G.edges(:,2) > nodes);
	if length(I) ~= 0
		valid = 0;
		if nargout >= 2
			reason = ['Edge indices not within 1-', num2str(nodes), '.'];
		end
		return;
	end

	% Check for self referential edges (self-loops)
	I = find (G.edges(:,1) == G.edges(:,2));
	if length(I) ~= 0
		valid = 0;
		if nargout >= 2
			reason = ['Graph has ', num2str(length(I)), ' self-loops.'];
		end
		return;
	end
end

% Check for duplicate edges
% DISABLED as both our gspan modification as well as mexrelabel support
%    multiple edge attributes
% TODO
if false
	E = G.edges(:,1:2);
	Eu = unique(E,'rows');
	if size(E,1) ~= size(Eu,1)
		valid = 0;
		if nargout >= 2
			reason = ['There are ', num2str(size(E,1)-size(Eu,1)), ...
				' duplicate edges.'];
		end
		return;
	end
end

if nargout >= 2
	reason = 'Graph is ok.';
end
valid = 1;

