/* mexgeosubkernel.cpp - Geometric subgraph kernel.
 *
 * Author: Sebastian Nowozin <sebastian.nowozin@tuebingen.mpg.de>
 * Date: 21th August 2007
 *
 * Usage:
 *
 * K = mexgeosubkernel (G, minsup, tolerance);
 *
 ***
 * Input
 *
 * G: (1,n) cellarray of n graph structures with this layout
 *    g.nodelabels: (n,1) discrete integer labels [L_1 ; L_2 ; ... ; L_n];
 *    g.nodepos: (n,3) real 3D coordinates of the vertex.
 *    g.edges: (m,2) edges, [from to] at each line:
 *       [e_1_{from} e_1_{to} edgelabel_1 ; ... ; e_m_{from} e_m_{to} edgelabel_m]
 *       The node indices go from 1 to n.
 * minsup: (optional) The minimum number of times a frequent subgraph has to
 *         appear in G.  This is useful only for transduction, otherwise the
 *         default of 1 should be used.
 * tolerance: (optional) Vertex match tolerance.  Default is 1e-1.
 *
 ***
 * Output
 *
 * K: (n,n) kernel matrix, encoding at (i,j) the number of common subgraphs
 *    sample i shares with sample j.
 */

#include <gmm/gmm.h>

#include <mex.h>

#include <mexhelp.h>
#include <Graph.h>
#include <FreqGeo.h>

class
MexGeoSubKRow : public FriGeo::FreqGeo
{
private:
	unsigned int count;
	std::vector<unsigned int> Krow;

protected:
	virtual void Report (const FriGeo::Graph& subg,
		const std::map<unsigned int,
			std::vector<std::vector<unsigned int> > >& L)
	{
		/* Don't consider the empty graph
		 */
		if (subg.VertexLabel ().size () == 0)
			return;

		count += 1;
		if (count % 10000 == 1)
			std::cout << "  " << count << " visited." << std::endl;

		for (std::map<unsigned int, std::vector<std::vector<unsigned int> > >::const_iterator
			li = L.begin () ; li != L.end () ; ++li)
		{
			Krow[li->first] += 1;	// += li->second.size ();
		}
	}

public:
	MexGeoSubKRow (const std::vector<FriGeo::Graph>& G,
		const FriGeo::Graph& refg, unsigned int minsup,
		double vertexTolerance)
		: FriGeo::FreqGeo (G, refg, minsup, vertexTolerance), count (0)
	{
		Krow.resize (G.size ());
	}

	const std::vector<unsigned int>& Row (void)
	{
		return (Krow);
	}
};


class
MexGeoSubK
{
private:
	gmm::dense_matrix<double> K;

	const std::vector<FriGeo::Graph>& G;
	unsigned int minsup;
	double vertexTolerance;

public:
	MexGeoSubK (const std::vector<FriGeo::Graph>& G, unsigned int minsup,
		double vertexTolerance)
		: G (G), minsup (minsup), vertexTolerance (vertexTolerance)
	{
	}

	void CalculateK (void)
	{
		K.resize (G.size (), G.size ());
		gmm::clear (K);

		for (unsigned int row = 0 ; row < G.size () ; ++row) {
			MexGeoSubKRow krow (G, G[row], minsup, vertexTolerance);
			std::cout << "Calculating row " << row << " ..." << std::endl;
			krow.Mine ();

			gmm::copy (gmm::row_vector (krow.Row ()),
				gmm::sub_matrix (K, gmm::sub_interval (row, 1),
					gmm::sub_interval (0, G.size ())));
		}
	}

	const gmm::dense_matrix<double>& GetK (void)
	{
		return (K);
	}
};


/* K = mexgeosubkernel (G, minsup, tolerance);
 */
void
mexFunction (int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[])
{
	if (nrhs < 1 || nrhs > 3) {
		mexErrMsgTxt ("Wrong number of right hand side parameters (must be 1, 2 or 3).\n");

		return;
	}

	if (nlhs != 1) {
		mexErrMsgTxt ("Wrong number of left hand side parameters (must be 1).\n");

		return;
	}

	if (mxIsCell (prhs[0]) == 0) {
		mexErrMsgTxt ("G must be a cellarray of graph structures.\n");

		return;
	}

	const mxArray* graphs = prhs[0];
	unsigned int gcount = mxGetN (graphs);
	unsigned int minsup = 1;
	if (nrhs >= 2)
		minsup = (unsigned int) mxGetScalar (prhs[1]);

	if (minsup > gcount) {
		mexErrMsgTxt ("The given minsup value is larger than the graph database size.\n");

		return;
	}

	const mxArray** garr = (const mxArray**) mxGetPr (graphs);
	std::vector<FriGeo::Graph> G;

	for (unsigned int n = 0 ; n < gcount ; ++n) {
		const mxArray* gcur = garr[n];

		FriGeo::Graph g;
		if (FriGeo::MexHelp::readGraph (gcur, g) == false)
			return;

		G.push_back (g);
	}

	double tolerance = 1e-1;
	if (nrhs >= 3)
		tolerance = mxGetScalar (prhs[2]);

	if (tolerance < 0.0) {
		mexErrMsgTxt ("The given matching tolerance is negative.\n");

		return;
	}

	MexGeoSubK gsk (G, minsup, tolerance);
	gsk.CalculateK ();

	plhs[0] = mxCreateNumericMatrix (G.size (), G.size (), mxDOUBLE_CLASS, 0);
	double* K = mxGetPr (plhs[0]);
	const gmm::dense_matrix<double>& Kcalc = gsk.GetK ();
	assert (G.size () == gmm::mat_nrows (Kcalc));
	assert (G.size () == gmm::mat_ncols (Kcalc));

	for (unsigned int row = 0 ; row < gmm::mat_nrows (Kcalc) ; ++row)
		for (unsigned int col = 0 ; col < gmm::mat_ncols (Kcalc) ; ++col)
			K[row+G.size()*col] = Kcalc(row,col);
}

