
#include <mexhelp.h>

namespace FriGeo
{

bool
MexHelp::readGraph (const mxArray* graph, FriGeo::Graph& gout)
{
	mxArray* nodelabels = mxGetField (graph, 0, "nodelabels");
	mxArray* nodepos = mxGetField (graph, 0, "nodepos");
	mxArray* edges = mxGetField (graph, 0, "edges");

	if (nodelabels == NULL || nodepos == NULL || edges == NULL) {
		mexErrMsgTxt ("Graph structures are missing the \"nodelabels\", "
			"\"nodepos\" or \"edges\" fields.\n");

		return (false);
	}

	if (mxIsDouble (nodepos) == 0) {
		mexErrMsgTxt ("\"nodepos\" not of type Double.\n");

		return (false);
	}

	if (mxIsUint32 (nodelabels) == 0 || mxIsUint32 (edges) == 0) {
		mexErrMsgTxt ("\"nodelabels\" or \"edges\" not of type Uint32.\n");

		return (false);
	}

	if (mxGetN (nodelabels) > 1) {
		mexErrMsgTxt ("\"nodelabels\" must be an Uint32 column vector.\n");

		return (false);
	}

	uint32_T* nmat = (uint32_T*) mxGetPr (nodelabels);
	assert (nmat != NULL);
	unsigned int nodes = mxGetM (nodelabels);
#ifdef	DEBUG
	mexPrintf ("Graph with %d node%s\n", nodes, nodes > 1 ? "s" : "");
#endif

	gout = FriGeo::Graph ();
	double* nodeposPr = mxGetPr (nodepos);
	for (unsigned int i = 0 ; i < nodes ; ++i) {
		std::vector<double> pos (3);

		pos[0] = nodeposPr[i+0*nodes];
		pos[1] = nodeposPr[i+1*nodes];
		pos[2] = nodeposPr[i+2*nodes];

		gout.PushVertex (pos, nmat[i]);
#ifdef	DEBUG
		mexPrintf ("inserting node %d with label %d\n", i, nmat[i]);
#endif
	}

	/* emat is column organised.
	 */
	unsigned int edgecount = mxGetM (edges);
	if (edgecount > 0) {
		uint32_T* emat = (uint32_T*) mxGetPr (edges);
#ifdef	DEBUG
		mexPrintf ("edges M, N: %d, %d\n", mxGetM (edges), mxGetN (edges));
#endif
		assert (emat != NULL);

		for (unsigned int i = 0 ; i < edgecount ; ++i) {
			unsigned int from = emat[i+0*edgecount] - 1;	// index correction
			unsigned int to = emat[i+1*edgecount] - 1;	// index correction
			unsigned int elabel = emat[i+2*edgecount];

#ifdef	DEBUG
			mexPrintf ("edge: from %d to %d, label %d\n", from, to, elabel);
#endif
			assert (from < nodes && to < nodes);
			if (from >= nodes || to >= nodes) {
				mexErrMsgTxt ("Edge from/to indices out of bounds.");

				return (false);
			}

			/* Warning: this makes edges undirected, which is not what we want (or
			 * do we?).
			 */
#ifdef	DEBUG
			mexPrintf ("inserting edge from %d to %d, label %d\n",
				from, to, elabel);
#endif
			gout.PushEdge (from, to, elabel);
		}
	}
	gout.Seal ();

	return (true);
}

mxArray*
MexHelp::writeGraph (const FriGeo::Graph& graph)
{
	const char* fieldnames[] = { "nodelabels", "nodepos", "edges", };
	mxArray* gs = mxCreateStructMatrix (1, 1, 3, fieldnames);
	assert (gs != NULL);

	/* First do the node labels.
	 */
	unsigned int nodecount = graph.VertexLabel ().size ();
	mxArray* nodelabels = mxCreateNumericMatrix
		(nodecount, 1, mxUINT32_CLASS, 0);
	assert (nodelabels != NULL);
	uint32_T* nodelabels_P = (uint32_T*) mxGetPr (nodelabels);
	assert (nodelabels_P != NULL);

	mxArray* nodepos = mxCreateNumericMatrix (nodecount, 3, mxDOUBLE_CLASS, 0);
	assert (nodepos != NULL);
	double* nodepos_P = mxGetPr (nodepos);
	assert (nodepos_P != NULL);

	for (int nidx = 0 ; nidx < nodecount ; ++nidx) {
		nodelabels_P[nidx] = graph.VertexLabel ()[nidx];

		nodepos_P[nidx+0*nodecount] = graph.VertexPosition ()(nidx, 0);
		nodepos_P[nidx+1*nodecount] = graph.VertexPosition ()(nidx, 1);
		nodepos_P[nidx+2*nodecount] = graph.VertexPosition ()(nidx, 2);
	}

	mxSetField (gs, 0, "nodelabels", nodelabels);
	mxSetField (gs, 0, "nodepos", nodepos);

	/* Now do the edges.
	 */
	unsigned int edgecount = gmm::mat_nrows (graph.Edges ());
	mxArray* edges = mxCreateNumericMatrix (edgecount, 3, mxUINT32_CLASS, 0);
	assert (edges != NULL);
	uint32_T* edges_P = (uint32_T*) mxGetPr (edges);
	assert (edges_P != NULL);

	for (unsigned int eidx = 0 ; eidx < edgecount ; ++eidx) {
		edges_P[eidx+0*edgecount] = graph.Edges ()(eidx,0) + 1;
		edges_P[eidx+1*edgecount] = graph.Edges ()(eidx,1) + 1;
		edges_P[eidx+2*edgecount] = graph.Edges ()(eidx,2);
	}
	mxSetField (gs, 0, "edges", edges);

	return (gs);
}

}

